using System.Net;
using DotNext;
namespace RemoteUpdate;

static class RemoteUpdateClientAddr
{

    /// <summary> 
    /// ADDR: 0X00: 写Flash-读写地址——控制位 <br/>
    /// [31:16]: wr_sector_num <br/>
    /// [15: 0]: {flash_wr_en,-,-,-, start_wr_sector} <br/>
    /// </summary>
    public const UInt32 WriteCtrl = 0x00;

    /// <summary> 
    /// ADDR: 0X01: 写Flash-只写地址——FIFO入口 <br/>
    /// [31:0]:  写比特流数据入口 <br/>
    /// </summary>
    public const UInt32 WriteFIFO = 0x01;

    /// <summary> 
    /// ADDR: 0X02: 写Flash-只读地址——标志位 <br/>
    /// [31:24]: {-, -, -, -, -, -, -,      wr_fifo_full} <br/>
    /// [23:16]: {-, -, -, -, -, -, -,     wr_fifo_empty} <br/>
    /// [15: 8]: {-, -, -, -, -, -, -,     flash_wr_done} <br/>
    /// [ 7: 0]: {-, -, -, -, -, -, -,  flash_clear_done} <br/>
    /// </summary>
    public const UInt32 WriteSign = 0x02;

    /// <summary> 
    /// ADDR: 0X03: 读Flash-读写地址——控制位1 <br/>
    /// [31:16]: rd_sector_num <br/>
    /// [15: 0]: {flash_rd_en,-,-,-, start_rd_sub_sector} <br/>
    /// </summary>
    public const UInt32 ReadCtrl1 = 0x03;

    /// <summary> 
    /// ADDR: 0X04: 读Flash-读写地址——控制位2 <br/>
    /// [31:24]: {                                            } <br/>
    /// [23:16]: {-, -, -, -, -, -,{   bs_crc32_ok           }} <br/>
    /// [15: 8]: {-, -, -, -, -, -, -,            crc_check_en} <br/>
    /// [ 7: 0]: {-, -, -, -, -, -, -,     bitstream_up2cpu_en} <br/>
    /// </summary>
    public const UInt32 ReadCtrl2 = 0x04;

    /// <summary> 
    /// ADDR: 0X05: 读Flash-只读地址——FIFO出口 <br/>
    /// [31:0]: 读比特流数据出口 <br/>
    /// </summary>
    public const UInt32 ReadFIFO = 0x05;

    /// <summary> 
    /// ADDR: 0X06: 读Flash-只读地址——CRC校验值 <br/>
    /// [31:0]: CRC校验值 bs_readback_crc <br/>
    /// </summary>
    public const UInt32 ReadCRC = 0x06;

    /// <summary> 
    /// ADDR: 0X07: 读Flash-只读地址——标志位 <br/>
    /// [31:24]: {-, -, -, -, -, -, -,         rd_fifo_afull} <br/>
    /// [23:16]: {-, -, -, -, -, -, -,         rd_fifo_empty} <br/>
    /// [15: 8]: {-, -, -, -, -, -, -,         flash_rd_done} <br/>
    /// [ 7: 0]: {-, -, -, -, -, -, -, bs_readback_crc_valid} <br/>
    /// </summary>
    public const UInt32 ReadSign = 0x07;

    /// <summary> 
    /// ADDR: 0X08: 热启动开关-读写地址——控制位 <br/>
    /// [31: 8]: hotreset_addr <br/>
    /// [ 7: 0]: {-, -, -, -, -, -, -,    hotreset_en} <br/>
    /// </summary>
    public const UInt32 HotResetCtrl = 0x08;

    /// <summary> 
    /// ADDR: 0X09: 只读地址 版本号 <br/>
    /// [31: 0]: FPGA_VERSION[31:0] <br/>
    /// </summary>
    public const UInt32 Version = 0x09;
}

static class FlashAddr
{
    public static readonly UInt32[] Switch = { 0x0000, 0x2000, 0x4000 };
    public static readonly UInt32[] Jump = { 0x1000, 0x3000, 0x5000 };
    public static readonly UInt32[] User = { 0x406000, 0x806000, 0xC06000 };

    public const UInt32 Golden = 0x6000;
}



/// <summary>
/// [TODO:description]
/// </summary>
public class RemoteUpdateClient
{
    private static readonly NLog.Logger logger = NLog.LogManager.GetCurrentClassLogger();

    const int FLASH_SECTOR_LENGTH = 4 * 1024;

    readonly int timeout = 2000;
    readonly int timeoutForWait = 60 * 1000;

    readonly int port;
    readonly string address;
    private IPEndPoint ep;

    /// <summary>
    /// [TODO:description]
    /// </summary>
    /// <param name="address">[TODO:parameter]</param>
    /// <param name="port">[TODO:parameter]</param>
    /// <param name="timeout">[TODO:parameter]</param>
    /// <param name="timeoutForWait">[TODO:parameter]</param>
    /// <returns>[TODO:return]</returns>
    public RemoteUpdateClient(string address, int port, int timeout = 2000, int timeoutForWait = 60 * 1000)
    {
        if (timeout < 0)
            throw new ArgumentException("Timeout couldn't be negative", nameof(timeout));
        this.address = address;
        this.port = port;
        this.ep = new IPEndPoint(IPAddress.Parse(address), port);
        this.timeout = timeout;
        this.timeoutForWait = timeoutForWait;
    }

    /// <summary>
    /// [TODO:description]
    /// </summary>
    /// <param name="flashAddr">[TODO:parameter]</param>
    /// <param name="writeSectorNum">[TODO:parameter]</param>
    /// <param name="bytesData">[TODO:parameter]</param>
    /// <returns>[TODO:return]</returns>
    public async ValueTask<Result<bool>> WriteFlash(UInt32 flashAddr, int writeSectorNum, byte[] bytesData)
    {
        // Assert
        if (writeSectorNum <= 0 || writeSectorNum >= FLASH_SECTOR_LENGTH)
            return new(new ArgumentException(
                $"Write sector num should be 1 ~ 4096, but given {writeSectorNum}", nameof(writeSectorNum)));
        if (bytesData.Length % (4 * 1024) != 0)
            return new(new ArgumentException(
                $"The length of data should be divided by 4096, bug given {bytesData.Length}", nameof(bytesData)));

        {
            var ret = await UDPClientPool.WriteAddr(
                this.ep, RemoteUpdateClientAddr.WriteCtrl,
                Convert.ToUInt32((writeSectorNum << 16) | (1 << 15) | Convert.ToInt32(flashAddr / 4096)), this.timeout);
            if (!ret.IsSuccessful) return new(ret.Error);
            if (!ret.Value) return new(new Exception("Enable write flash failed"));
        }

        {
            var ret = await UDPClientPool.ReadAddrWithWait(
                this.ep, RemoteUpdateClientAddr.WriteSign,
                0x00_00_00_01, 0x00_00_00_01, this.timeoutForWait);
            if (!ret.IsSuccessful) return new(ret.Error);
            if (!ret.Value) return new(new Exception(
                $"Flash clear failed after {this.timeoutForWait} milliseconds"));
        }

        {
            var ret = await UDPClientPool.WriteAddr(this.ep, RemoteUpdateClientAddr.WriteFIFO, bytesData, this.timeout);
            if (!ret.IsSuccessful) return new(ret.Error);
            if (!ret.Value) return new(new Exception("Send data to flash failed"));
        }

        {
            var ret = await UDPClientPool.ReadAddrWithWait(
                this.ep, RemoteUpdateClientAddr.WriteSign,
                0x00_00_01_00, 0x00_00_01_00, this.timeoutForWait);
            if (!ret.IsSuccessful) return new(ret.Error);
            return ret.Value;
        }
    }

    /// <summary>
    /// [TODO:description]
    /// </summary>
    /// <param name="bitstreamNum">[TODO:parameter]</param>
    /// <returns>[TODO:return]</returns>
    public async ValueTask<Result<bool>> EnableBitstream(int bitstreamNum)
    {
        // Assert
        if (bitstreamNum <= 0 || bitstreamNum > 3)
            return new(new ArgumentException(
                $"Bitsteam num should be 1 ~ 3, but given {bitstreamNum}", nameof(bitstreamNum)));

        var bytesData = new byte[FLASH_SECTOR_LENGTH];
        Array.Fill<byte>(bytesData, 0xFF);
        bytesData[FLASH_SECTOR_LENGTH - 1] = 0x01;
        bytesData[FLASH_SECTOR_LENGTH - 2] = 0x33;
        bytesData[FLASH_SECTOR_LENGTH - 3] = 0x2D;
        bytesData[FLASH_SECTOR_LENGTH - 4] = 0x94;

        var ret = await WriteFlash(FlashAddr.Switch[bitstreamNum], 1, bytesData);
        if (!ret.IsSuccessful) return new(ret.Error);
        return ret.Value;
    }

    /// <summary>
    /// [TODO:description]
    /// </summary>
    /// <param name="bitstreamNum">[TODO:parameter]</param>
    /// <returns>[TODO:return]</returns>
    public async ValueTask<Result<bool>> DisableBitstream(int bitstreamNum)
    {
        // Assert
        if (bitstreamNum <= 0 || bitstreamNum > 3)
            return new(new ArgumentException(
                $"Bitsteam num should be 1 ~ 3, but given {bitstreamNum}", nameof(bitstreamNum)));

        var bytesData = new byte[FLASH_SECTOR_LENGTH];
        Array.Fill<byte>(bytesData, 0xFF);

        var ret = await WriteFlash(FlashAddr.Switch[bitstreamNum], 1, bytesData);
        if (!ret.IsSuccessful) return new(ret.Error);
        return ret.Value;
    }


    /// <summary>
    /// [TODO:description]
    /// </summary>
    /// <param name="bitstreamNum">[TODO:parameter]</param>
    /// <returns>[TODO:return]</returns>
    public async ValueTask<Result<bool>> WriteJumpCmd(int bitstreamNum)
    {
        // Assert
        if (bitstreamNum <= 0 || bitstreamNum > 3)
            return new(new ArgumentException(
                $"Bitsteam num should be 1 ~ 3, but given {bitstreamNum}", nameof(bitstreamNum)));

        // Init data
        var bytesData = new byte[FLASH_SECTOR_LENGTH];
        for (int i = 0; i < FLASH_SECTOR_LENGTH; i += 4)
            bytesData[i] = 0xA0;

        Buffer.BlockCopy(new byte[] { 0xAB, 0x00, 0x00, 0x01 }, 0, bytesData, 0x04, 4);
        Buffer.BlockCopy(new byte[] { 0x00, 0x00, 0x00, 0x0B }, 0, bytesData, 0x08, 4);
        Buffer.BlockCopy(new byte[] { 0xAB, 0xC0, 0x00, 0x01 }, 0, bytesData, 0x34, 4);
        Buffer.BlockCopy(new byte[] { 0x00, 0x00, 0x00, 0x00 }, 0, bytesData, 0x38, 4);
        Buffer.BlockCopy(new byte[] { 0xAC, 0x00, 0x00, 0x01 }, 0, bytesData, 0x3C, 4);
        Buffer.BlockCopy(Common.Number.NumberToBytes(FlashAddr.User[bitstreamNum], 4).Value, 0, bytesData, 0x40, 4);
        Buffer.BlockCopy(new byte[] { 0xA8, 0x80, 0x00, 0x01 }, 0, bytesData, 0x44, 4);
        Buffer.BlockCopy(new byte[] { 0x00, 0x00, 0x00, 0x0F }, 0, bytesData, 0x48, 4);

        var ret = await WriteFlash(FlashAddr.Switch[bitstreamNum], 1, bytesData);
        if (!ret.IsSuccessful) return new(ret.Error);
        return ret.Value;
    }

    /// <summary>
    /// [TODO:description]
    /// </summary>
    /// <param name="bitstreamNum">[TODO:parameter]</param>
    /// <returns>[TODO:return]</returns>
    public async ValueTask<Result<bool>> InitRemoteUpdate(int bitstreamNum)
    {
        // Assert
        if (bitstreamNum <= 0 || bitstreamNum > 3)
            return new(new ArgumentException(
                $"Bitsteam num should be 1 ~ 3, but given {bitstreamNum}", nameof(bitstreamNum)));

        for (int i = 0; i < 3; i++)
        {
            {
                var ret = (i == bitstreamNum) ? await EnableBitstream(i) : await DisableBitstream(i);
                if (!ret.IsSuccessful) return new(ret.Error);
                if (!ret.Value) return new(new Exception($"Write switch {i} failed"));
            }

            {
                var ret = await WriteJumpCmd(i);
                if (!ret.IsSuccessful) return new(ret.Error);
                if (!ret.Value) return new(new Exception($"Write jump {i} failed"));
            }
        }

        return true;
    }


    /// <summary>
    /// [TODO:description]
    /// </summary>
    /// <param name="bitstreamNum">[TODO:parameter]</param>
    /// <param name="checkSum">[TODO:parameter]</param>
    /// <returns>[TODO:return]</returns>
    public async ValueTask<Result<bool>> CheckBitstreamCRC(int bitstreamNum, UInt32 checkSum)
    {
        {
            var ret = await UDPClientPool.WriteAddr(this.ep, RemoteUpdateClientAddr.ReadCtrl2, 0x00_00_00_00, this.timeout);
            if (!ret.IsSuccessful) return new(ret.Error);
            if (!ret.Value) return new(new Exception("Write read control 2 failed"));
        }

        {
            var ret = await UDPClientPool.WriteAddr(
                this.ep, RemoteUpdateClientAddr.ReadCtrl1,
                Convert.ToUInt32((1024 << 16) | (1 << 15) | Convert.ToInt32(FlashAddr.User[bitstreamNum] / 4096)), this.timeout);
            if (!ret.IsSuccessful) return new(ret.Error);
            if (!ret.Value) return new(new Exception("Write read control 1 failed"));
        }

        {
            var ret = await UDPClientPool.ReadAddrWithWait(
                this.ep, RemoteUpdateClientAddr.ReadSign,
                0x00_00_01_00, 0x00_00_01_00, this.timeoutForWait);
            if (!ret.IsSuccessful) return new(ret.Error);
            if (!ret.Value) return new(new Exception(
                $"Read bitstream failed after {this.timeoutForWait} milliseconds"));
        }

        {
            var ret = await UDPClientPool.ReadAddr(this.ep, RemoteUpdateClientAddr.ReadCRC, this.timeout);
            if (!ret.IsSuccessful) return new(ret.Error);

            var bytes = ret.Value.Options.Data;
            if (bytes is null)
                return new(new Exception("CRC is null"));

            var remoteCRC = Common.Number.BytesToUInt32(bytes);
            if (!remoteCRC.IsSuccessful) return new(remoteCRC.Error);

            logger.Debug($"Expected CRC: \t 0x{Convert.ToString(checkSum, 16)}");
            logger.Debug($"Received CRC: \t 0x{Convert.ToString(remoteCRC.Value, 16)}");

            return remoteCRC.Value == checkSum;
        }
    }

    /// <summary>
    /// [TODO:description]
    /// </summary>
    /// <param name="bitstreamNum">[TODO:parameter]</param>
    /// <returns>[TODO:return]</returns>
    public async ValueTask<Result<bool>> HotRest(int bitstreamNum)
    {
        // Assert
        if (bitstreamNum <= 0 || bitstreamNum > 3)
            return new(new ArgumentException(
                $"Bitsteam num should be 1 ~ 3, but given {bitstreamNum}", nameof(bitstreamNum)));

        var ret = await UDPClientPool.WriteAddr(
            this.ep, RemoteUpdateClientAddr.HotResetCtrl,
            ((FlashAddr.User[bitstreamNum] << 8) | 1), this.timeout);
        if (!ret.IsSuccessful) return new(ret.Error);
        return ret.Value;
    }

    /// <summary>
    /// [TODO:description]
    /// </summary>
    /// <param name="bitstreamNum">[TODO:parameter]</param>
    /// <param name="bytesData">[TODO:parameter]</param>
    /// <returns>[TODO:return]</returns>
    public async ValueTask<Result<bool>> UpdateBitstream(int bitstreamNum, byte[] bytesData)
    {
        // Assert
        if (bitstreamNum <= 0 || bitstreamNum > 3)
            return new(new ArgumentException(
                $"Bitsteam num should be 1 ~ 3, but given {bitstreamNum}", nameof(bitstreamNum)));

        {
            var ret = await InitRemoteUpdate(bitstreamNum);
            if (!ret.IsSuccessful) return new(ret.Error);
            if (!ret.Value) return new(new Exception("Init remote update failed"));
        }

        {
            var ret = await WriteFlash(FlashAddr.User[bitstreamNum], 1024, bytesData);
            if (!ret.IsSuccessful) return new(ret.Error);
            if (!ret.Value) return new(new Exception("Write bitstream failed"));
        }

        {
            var ret = await HotRest(bitstreamNum);
            if (!ret.IsSuccessful) return new(ret.Error);
            return ret.Value;
        }
    }
}
